"""Classes and methods related to the Structure Notation Language (SNL)."""

from __future__ import annotations

import json
import re
import sys
from datetime import UTC, datetime
from typing import TYPE_CHECKING, NamedTuple

from monty.json import MontyDecoder, MontyEncoder

from pymatgen.core.structure import Molecule, Structure

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Self


__author__ = "Anubhav Jain, Shyue Ping Ong"
__credits__ = "Dan Gunter"


MAX_HNODE_SIZE: int = 64_000  # maximum size (bytes) of SNL HistoryNode
MAX_DATA_SIZE: int = 256_000  # maximum size (bytes) of SNL data field
MAX_HNODES: int = 100  # maximum number of HistoryNodes in SNL file
MAX_BIBTEX_CHARS: int = 20_000  # maximum number of characters for BibTeX reference


def is_valid_bibtex(reference: str) -> bool:
    """Validate that a reference is in proper BibTeX format.

    Args:
        reference (str): Reference in BibTeX format.

    Returns:
        bool: True if reference is valid BibTeX.
    """
    from bibtexparser.bparser import BibTexParser

    parser = BibTexParser()
    try:
        bib_database = parser.parse(reference)
        return bool(bib_database.entries)
    except Exception:
        return False


class HistoryNode(NamedTuple):
    """A HistoryNode represents a step in the chain of events that lead to a
    Structure. HistoryNodes leave 'breadcrumbs' so that you can trace back how
    a Structure was created. For example, a HistoryNode might represent pulling
    a Structure from an external database such as the ICSD or CSD. Or, it might
    represent the application of a code (e.g. pymatgen) to the Structure, with
    a custom description of how that code was applied (e.g. a site removal
    Transformation was applied).

    Attributes:
        name (str): The name of a code or resource that this Structure encountered in its history.
        url (str): The URL of that code/resource.
        description (str): A free-form description of how the code/resource is related to the Structure.
    """

    name: str
    url: str
    description: str

    def as_dict(self) -> dict[str, str]:
        """Get MSONable dict."""
        return {"name": self.name, "url": self.url, "description": self.description}

    @classmethod
    def from_dict(cls, dct: dict[str, str]) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            HistoryNode
        """
        return cls(dct["name"], dct["url"], dct["description"])

    @classmethod
    def parse_history_node(cls, h_node) -> Self:
        """Parse a History Node object from either a dict or a tuple.

        Args:
            h_node: A dict with name/url/description fields or a 3-element tuple.

        Returns:
            HistoryNode
        """
        if isinstance(h_node, dict):
            return cls.from_dict(h_node)

        if len(h_node) != 3:
            raise ValueError(f"Invalid History node, should be dict or (name, version, description) tuple: {h_node}")
        return cls(h_node[0], h_node[1], h_node[2])


class Author(NamedTuple):
    """An Author contains two fields: name and email. It is meant to represent
    the author of a Structure or the author of a code that was applied to a Structure.
    """

    name: str
    email: str

    def __str__(self) -> str:
        """String representation of an Author."""
        return f"{self.name} <{self.email}>"

    def as_dict(self) -> dict[str, str]:
        """Get MSONable dict."""
        return {"name": self.name, "email": self.email}

    @classmethod
    def from_dict(cls, dct: dict[str, str]) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            Author
        """
        return cls(dct["name"], dct["email"])

    @classmethod
    def parse_author(
        cls,
        author: str | dict[str, str] | tuple[str, str],
    ) -> Self:
        """Parse an Author object from either a String, dict, or tuple.

        Args:
            author: in one of the three accepted formats:
                - A string formatted as "NAME <email@domain.com>".
                - A (name, email) tuple.
                - A dict with name and email as keys.

        Returns:
            An Author object.
        """
        if isinstance(author, str):
            # Regex looks for whitespace, (any name), whitespace, <, (email),
            # >, whitespace
            match = re.match(r"\s*(.*?)\s*<(.*?@.*?)>\s*", author)
            if not match or match.start() != 0 or match.end() != len(author):
                raise ValueError(f"Invalid author format! {author}")
            return cls(match.groups()[0], match.groups()[1])

        if isinstance(author, dict):
            return cls.from_dict(author)

        if len(author) != 2:
            raise ValueError(f"Invalid author, should be String or (name, email) tuple: {author}")

        return cls(author[0], author[1])


class StructureNL:
    """The Structure Notation Language (SNL, pronounced "snail") is a container for a pymatgen
    Structure/Molecule object with some additional fields for enhanced provenance.

    It is meant to be imported/exported in a JSON file format with the following structure:
        - sites
        - lattice (optional)
        - about
            - created_at
            - authors
            - projects
            - references
            - remarks
            - data
            - history
    """

    def __init__(
        self,
        struct_or_mol: Structure | Molecule,
        authors: str | list[dict] | list[str],
        projects: list[str] | str | None = None,
        references: str = "",
        remarks: list[str] | str | None = None,
        data: dict | None = None,
        history: list[dict] | None = None,
        created_at: datetime | None = None,
    ) -> None:
        """
        Args:
            struct_or_mol: A pymatgen Structure/Molecule object
            authors (list | str):
                - List of {"name":"", "email":""} dicts.
                - List of strings as "John Doe <johndoe@gmail.com>".
                - A single String with commas separating authors.
            projects (list[str] | str): e.g. ["Project A", "Project B"]
            references (str): A string in BibTeX format
            remarks (list[str]): e.g. ["Remark A", "Remark B"]
            data (dict): A free form dict. Namespaced at the root level with an
                underscore, e.g. {"_materialsproject": <custom data>}
            history (list[dict]): e.g. [{"name":"", "url":"", "description":{}}]
            created_at (datetime): Creation date.
        """
        # Initialize root-level structure keys
        self.structure = struct_or_mol

        # Turn `authors` into list of `Author` objects
        _authors: list = authors.split(",") if isinstance(authors, str) else authors
        self.authors: list[Author] = [Author.parse_author(a) for a in _authors]

        # Turn `projects` into list of strings
        projects = projects or []
        self.projects: list[str] = [projects] if isinstance(projects, str) else projects

        # Check that references are valid BibTeX string
        if not isinstance(references, str):
            raise TypeError("Invalid format for SNL reference! Should be empty string or BibTeX string.")

        if references and not is_valid_bibtex(references):
            raise ValueError("Invalid format for SNL reference! Should be BibTeX string.")

        if len(references) > MAX_BIBTEX_CHARS:
            raise ValueError(
                f"The BibTeX string must be fewer than {MAX_BIBTEX_CHARS} chars, you have {len(references)}"
            )

        self.references = references

        # Turn `remarks` into list of strings
        remarks = remarks or []
        self.remarks: list[str] = [remarks] if isinstance(remarks, str) else remarks

        # Check `remarks` length limit
        for remark in self.remarks:
            if len(remark) > 140:
                raise ValueError(f"The remark exceeds the maximum size of 140 characters: {len(remark)}")

        # Check data length limit
        self.data = data or {}
        if not sys.getsizeof(self.data) < MAX_DATA_SIZE:
            raise ValueError(
                f"The data dict exceeds the maximum size limit of {MAX_DATA_SIZE} "
                f"bytes (you have {sys.getsizeof(data)})"
            )

        for key in self.data:
            if not key.startswith("_"):
                raise ValueError(
                    "data must contain properly namespaced data with keys starting with an underscore. "
                    f"{key=} does not start with an underscore."
                )

        # Check for valid history nodes
        history = history or []  # initialize null fields
        if len(history) > MAX_HNODES:
            raise ValueError(f"A maximum of {MAX_HNODES} History nodes are supported, you have {len(history)}!")
        self.history = [HistoryNode.parse_history_node(h) for h in history]
        if not all(sys.getsizeof(h) < MAX_HNODE_SIZE for h in history):
            raise ValueError(f"One or more history nodes exceeds the maximum size limit of {MAX_HNODE_SIZE} bytes")

        self.created_at = created_at or f"{datetime.now(tz=UTC):%Y-%m-%d %H:%M:%S.%f%z}"

    def __str__(self) -> str:
        return "\n".join(
            [
                f"{key}\n{getattr(self, key)}"
                for key in (
                    "structure",
                    "authors",
                    "projects",
                    "references",
                    "remarks",
                    "data",
                    "history",
                    "created_at",
                )
            ]
        )

    def __eq__(self, other: object) -> bool:
        """Check for equality between two StructureNL objects."""
        needed_attrs = (
            "structure",
            "authors",
            "projects",
            "references",
            "remarks",
            "data",
            "history",
            "created_at",
        )

        if not all(hasattr(other, attr) for attr in needed_attrs):
            return NotImplemented

        return all(getattr(self, attr) == getattr(other, attr) for attr in needed_attrs)

    def as_dict(self) -> dict[str, Any]:
        """Get MSONable dict."""
        dct = self.structure.as_dict()
        dct["@module"] = type(self).__module__
        dct["@class"] = type(self).__name__
        dct["about"] = {
            "authors": [a.as_dict() for a in self.authors],
            "projects": self.projects,
            "references": self.references,
            "remarks": self.remarks,
            "history": [h.as_dict() for h in self.history],
            "created_at": json.loads(json.dumps(self.created_at, cls=MontyEncoder)),
        }
        dct["about"].update(json.loads(json.dumps(self.data, cls=MontyEncoder)))
        return dct

    @classmethod
    def from_dict(cls, dct: dict[str, Any]) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            Class
        """
        about = dct["about"]

        created_at = MontyDecoder().process_decoded(about.get("created_at"))
        data = {k: v for k, v in dct["about"].items() if k.startswith("_")}
        data = MontyDecoder().process_decoded(data)

        structure = Structure.from_dict(dct) if "lattice" in dct else Molecule.from_dict(dct)
        return cls(
            structure,
            about["authors"],
            projects=about.get("projects"),
            references=about.get("references", ""),
            remarks=about.get("remarks"),
            data=data,
            history=about.get("history"),
            created_at=created_at,
        )

    @classmethod
    def from_structures(
        cls,
        structures: Sequence[Structure],
        authors,
        projects: list[str] | None = None,
        references: str = "",
        remarks: list[str] | None = None,
        data: list[dict] | None = None,
        histories: list[list[dict]] | None = None,
        created_at: datetime | None = None,
    ) -> list[Self]:
        """A convenience method for getting a list of StructureNL objects by
        specifying structures and metadata separately. Some of the metadata
        is applied to all of the structures for ease of use.

        Args:
            structures (Sequence[Structure]): Structure objects
            authors: *List* of {"name":'', "email":''} dicts,
                *list* of Strings as 'John Doe <johndoe@gmail.com>',
                or a single String with commas separating authors
            projects (list[str]): e.g. ["Project A", "Project B"]. This
                applies to all structures.
            references (str): A String in BibTeX format. This applies to all
                structures.
            remarks (list[str]): e.g. ["Remark A", "Remark B"].
            data (list[dict]): Free form dicts. Namespaced at the root level
                with an underscore, e.g. {"_materialsproject":<custom data>}
                . The length of data should be the same as the list of
                structures if not None.
            histories(list[list[dict]]): e.g. [[{"name":"", "url":"",
                "description":{}}], ...] The length of histories should be the
                same as the list of structures if not None.
            created_at (datetime): Creation date.
        """
        data = [{}] * len(structures) if data is None else data
        histories = [[]] * len(structures) if histories is None else histories

        snl_list = []
        for idx, struct in enumerate(structures):
            snl = cls(
                struct,
                authors,
                projects=projects,
                references=references,
                remarks=remarks,
                data=data[idx],
                history=histories[idx],
                created_at=created_at,
            )
            snl_list.append(snl)

        return snl_list



"""Get help with VASP parameters from VASP wiki."""

from __future__ import annotations

import re

import orjson
import requests


class VaspDoc:
    """A VASP documentation helper."""

    def __init__(self) -> None:
        """Init for VaspDoc."""
        self.url_template = "https://www.vasp.at/wiki/index.php/%s"

    def print_help(self, tag: str) -> None:
        """
        Print the help for a TAG.

        Args:
            tag (str): Tag used in VASP.
        """
        print(self.get_help(tag))

    def print_jupyter_help(self, tag: str) -> None:
        """
        Display HTML help in ipython notebook.

        Args:
            tag (str): Tag used in VASP.
        """
        html_str = self.get_help(tag, "html")
        from IPython.core.display import HTML, display

        display(HTML(html_str))

    @classmethod
    def get_help(cls, tag: str, fmt: str = "text") -> str:
        """Get help on a VASP tag.

        Args:
            tag (str): VASP tag, e.g. ISYM.

        Returns:
            Help text.
        """
        try:
            from bs4 import BeautifulSoup
        except ImportError as exc:
            raise ImportError("BeautifulSoup4 must be installed to fetch from the VASP wiki.") from exc

        tag = tag.upper()
        response = requests.get(
            f"https://www.vasp.at/wiki/index.php/{tag}",
            timeout=60,
        )
        response.raise_for_status()

        soup = BeautifulSoup(response.text, features="html.parser")
        main_doc = soup.find(id="mw-content-text")
        if fmt == "text":
            output = main_doc.text
            return re.sub("\n{2,}", "\n\n", output)

        return str(main_doc)

    @classmethod
    def get_incar_tags(cls) -> list[str]:
        """Get a list of all INCAR tags from the VASP wiki."""
        # Use Mediawiki API as documented in
        # https://www.vasp.at/wiki/api.php?action=help&modules=query
        url = (
            "https://www.vasp.at/wiki/api.php?"
            "action=query&list=categorymembers"
            "&cmtitle=Category:INCAR_tag"
            "&cmlimit=500&format=json"
        )
        response = requests.get(url, timeout=60)
        response.raise_for_status()

        response_dict = orjson.loads(response.text)

        def extract_titles(data):
            """Extract keywords from from Wikimedia response data.
            See https://www.vasp.at/wiki/api.php?action=help&modules=query%2Bcategorymembers
            Returns: List of keywords as strings.
            """
            return [category_data["title"] for category_data in data["query"]["categorymembers"]]

        tags = extract_titles(response_dict)

        # If there are more than 500 items in the response, we will
        # get 'continue' field in the response
        # See https://www.mediawiki.org/wiki/API:Continue
        while "continue" in response_dict:
            response = requests.get(url + f"&cmcontinue={response_dict['continue']['cmcontinue']}", timeout=60)
            response.raise_for_status()

            response_dict = orjson.loads(response.text)
            tags = tags + extract_titles(response_dict)

        return tags



"""This module implements symmetry-related structure forms."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from monty.json import MontyEncoder
from tabulate import tabulate

from pymatgen.core.structure import FileFormats, PeriodicSite, Structure

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Self

    from numpy.typing import ArrayLike

    from pymatgen.core import IStructure
    from pymatgen.symmetry.analyzer import SpacegroupOperations
    from pymatgen.util.typing import PathLike


class SymmetrizedStructure(Structure):
    """This class represents a symmetrized structure, i.e. a structure
    where the space group and symmetry operations are defined. This class is
    typically not called but instead is typically obtained by calling
    pymatgen.symmetry.analyzer.SpacegroupAnalyzer.get_symmetrized_structure.

    Attributes:
        equivalent_indices (list[list[int]]): Indices of the sites in the structure that are
            considered equivalent based on the symmetry operations of the space group.
    """

    def __init__(
        self,
        structure: Structure | IStructure,
        spacegroup: SpacegroupOperations,
        equivalent_positions: ArrayLike,
        wyckoff_letters: Sequence[str],
    ) -> None:
        """
        Args:
            structure (Structure): Original structure
            spacegroup (SpacegroupOperations): An input SpacegroupOperations from SpacegroupAnalyzer.
            equivalent_positions (list[int]): Equivalent positions from SpacegroupAnalyzer.
            wyckoff_letters (list[str]): Wyckoff letters.
        """
        self.spacegroup = spacegroup
        uniq, inverse = np.unique(equivalent_positions, return_inverse=True)
        inverse = np.atleast_1d(inverse)  # Ensures `inverse` is 1D
        self.site_labels = equivalent_positions

        super().__init__(
            structure.lattice,
            [site.species for site in structure],
            structure.frac_coords,
            site_properties=structure.site_properties,
            properties=structure.properties,
            labels=structure.labels,
        )

        equivalent_indices: list[list[int]] = [[] for _ in range(len(uniq))]
        equivalent_sites: list[list[PeriodicSite]] = [[] for _ in range(len(uniq))]
        wyckoff_symbols: list[list[str]] = [[] for _ in range(len(uniq))]
        for idx, inv_ in enumerate(inverse):
            equivalent_indices[inv_].append(idx)
            equivalent_sites[inv_].append(self.sites[idx])
            wyckoff_symbols[inv_].append(wyckoff_letters[idx])
        self.equivalent_indices = equivalent_indices
        self.equivalent_sites = equivalent_sites
        self.wyckoff_letters = wyckoff_letters
        self.wyckoff_symbols = [f"{len(symb)}{symb[0]}" for symb in wyckoff_symbols]

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        outs = [
            "SymmetrizedStructure",
            f"Full Formula ({self.formula})",
            f"Reduced Formula: {self.reduced_formula}",
            f"Spacegroup: {self.spacegroup.int_symbol} ({self.spacegroup.int_number})",
            f"abc   : {' '.join(f'{val:>10.6f}' for val in self.lattice.abc)}",
            f"angles: {' '.join(f'{val:>10.6f}' for val in self.lattice.angles)}",
        ]

        if self._charge:
            outs.append(f"Overall Charge: {self._charge:+}")
        outs.append(f"Sites ({len(self)})")
        data = []
        props = self.site_properties
        keys = sorted(props)
        for idx, sites in enumerate(self.equivalent_sites):
            site = sites[0]
            row = [str(idx), site.species_string]
            row.extend([f"{j:>10.6f}" for j in site.frac_coords])
            row.append(self.wyckoff_symbols[idx])
            row += [props[key][idx] for key in keys]
            data.append(row)
        outs.append(tabulate(data, headers=["#", "SP", "a", "b", "c", "Wyckoff", *keys]))
        return "\n".join(outs)

    def copy(self) -> Self:
        """Make a copy of the SymmetrizedStructure."""
        return type(self)(
            self,
            spacegroup=self.spacegroup,
            equivalent_positions=self.site_labels,
            wyckoff_letters=self.wyckoff_letters,
        )

    def find_equivalent_sites(self, site: PeriodicSite) -> list[PeriodicSite]:
        """Find all symmetrically equivalent sites for a particular site.

        Args:
            site (PeriodicSite): A site in the structure.

        Raises:
            ValueError: if site is not in the structure.

        Returns:
            list[PeriodicSite]: all symmetrically equivalent sites.
        """
        for sites in self.equivalent_sites:
            if site in sites:
                return sites

        raise ValueError("Site not in structure")

    def as_dict(self) -> dict[str, Any]:
        """MSONable dict."""
        structure = Structure.from_sites(self.sites)
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "structure": structure.as_dict(),
            "spacegroup": self.spacegroup,
            "equivalent_positions": self.site_labels,
            "wyckoff_letters": self.wyckoff_letters,
        }

    @classmethod
    def from_dict(cls, dct: dict[str, Any]) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            SymmetrizedStructure
        """
        return cls(
            Structure.from_dict(dct["structure"]),
            spacegroup=dct["spacegroup"],
            equivalent_positions=dct["equivalent_positions"],
            wyckoff_letters=dct["wyckoff_letters"],
        )

    def to(self, filename: PathLike = "", fmt: FileFormats = "", **kwargs) -> str:
        """Use `MontyEncoder` as default JSON encoder."""
        filename, fmt = str(filename), cast("FileFormats", fmt.lower())
        if fmt == "json":
            kwargs.setdefault("cls", MontyEncoder)

        return super().to(filename, fmt, **kwargs)



"""This package implements modules for input and output to and from PWmat."""

from __future__ import annotations

from .inputs import AtomConfig
from .outputs import DosSpin, Movement, OutFermi, Report



from __future__ import annotations

from abc import abstractmethod
from collections import Counter
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Self, cast

import numpy as np
from monty.io import zopen
from monty.json import MontyDecoder, MSONable

from pymatgen.io.lobster.future.constants import LOBSTER_VERSION
from pymatgen.io.lobster.future.utils import convert_spin_keys, restore_spin_keys

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from numpy.typing import NDArray

    from pymatgen.io.lobster.future.types import LobsterInteraction, LobsterInteractionData, Spin
    from pymatgen.util.typing import PathLike


class LobsterFile(MSONable):
    """
    Representation of a LOBSTER file.

    This class provides a framework for parsing and processing LOBSTER output files.
    It supports version-specific processing through a registry of version processors.

    Attributes:
        filename (PathLike): Name or path of the file. Defaults to the class default.
        lobster_version (str): Version string parsed from the file or the default LOBSTER_VERSION.
        version_processors (ClassVar[dict[tuple[str, str | None], Callable]]): Registry of version processors.
        spins (list[Spin] | None): List of Spin objects if the file contains spin-polarized data, else None.
    """

    version_processors: ClassVar[dict[tuple[str, str | None], Callable]]

    spins: list[Spin] | None = None

    def __init__(
        self,
        filename: PathLike | None = None,
        process_immediately: bool = True,
        lobster_version: str | None = None,
    ) -> None:
        """
        Initialize a LobsterFile instance.

        Args:
            filename (PathLike | None): Path to the file. If None, uses the default filename.
            process_immediately (bool): Whether to process the file immediately upon initialization. Defaults to True.
            lobster_version (str | None): LOBSTER version string to use for parsing. If None, attempts to detect
            from file or falls back to default.
        """
        self.filename = Path(filename or self.get_default_filename()).expanduser().resolve()
        self.lobster_version = lobster_version or self.get_file_version() or LOBSTER_VERSION

        if process_immediately:
            self.process()

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Automatically registers version processors for subclasses.
        This method scans the subclass for methods decorated with @version_processor
        and adds them to the version_processors registry.
        """
        super().__init_subclass__(**kwargs)

        version_processors = {}
        for base in cls.__bases__:
            if hasattr(base, "version_processors"):
                version_processors.update(base.version_processors)

        for value in cls.__dict__.values():
            if hasattr(value, "version_info"):
                version_processors[value.version_info] = value

        cls.version_processors = version_processors

    def process(self) -> None:
        """
        Processes the file using the appropriate version processor.

        Selects the best matching version processor from the `version_processors` registry and invokes it.

        Raises:
            ValueError: If no processor matches the file version.
            RuntimeError: If the selected processor raises an exception.
        """
        eligible_methods = []

        if not self.filename.exists():
            raise FileNotFoundError(f"The file {self.filename} does not exist.")

        for (
            minimum_version,
            maximum_version,
        ), processor in self.version_processors.items():
            if LobsterFile.check_version(self.lobster_version, minimum_version, maximum_version):
                eligible_methods.append((minimum_version, processor))

        if not eligible_methods:
            raise ValueError(f"No processor found for version {self.lobster_version}")

        best_method = max(eligible_methods, key=lambda x: x[0])[-1]

        try:
            best_method(self)
        except Exception as e:
            processor_name = getattr(best_method, "__name__", str(best_method))

            raise RuntimeError(
                f"Error occurred during file processing with {processor_name} (version {self.lobster_version}): {e}"
            ) from e

    @staticmethod
    def check_version(actual: str, minimum: str, maximum: str | None) -> bool:
        """
        Checks whether a version string falls within a min/max inclusive range.

        Args:
            actual (str): Version string to check (e.g., "1.2.3").
            minimum (str): Minimum acceptable version string (exclusive).
            maximum (str | None): Maximum acceptable version string (exclusive) or None for no upper bound.

        Returns:
            bool: True if `actual` is > `minimum` and < `maximum` (if provided), otherwise False.
        """
        actual_parts = tuple(map(int, actual.split(".")))
        minimum_parts = tuple(map(int, minimum.split(".")))

        if actual_parts < minimum_parts:
            return False

        if maximum is not None:
            maximum_parts = tuple(map(int, maximum.split(".")))

            if actual_parts > maximum_parts:
                return False

        return True

    def get_file_version(self) -> str | None:
        """
        Retrieves the file version. Override in subclasses to extract version from file content if possible.

        Returns:
            str | None: Version string (e.g., "1.2.3") if found, else None.
        """

    @classmethod
    @abstractmethod
    def get_default_filename(cls) -> str:
        """
        Returns the default filename for this LobsterFile subclass.

        Returns:
            str: The default filename.
        """
        ...

    @cached_property
    def lines(self) -> list[str]:
        """
        Returns all lines from the file as a list of strings.

        Returns:
            list[str]: Lines from the file.

        Raises:
            ValueError: If the file is empty.
        """
        with zopen(self.filename, mode="rt", encoding="utf-8") as file:
            lines = file.read().splitlines()

            if len(lines) == 0:
                raise ValueError(f"The file {self.filename} is empty.")

            return cast("list[str]", lines)

    def iterate_lines(self) -> Generator[str]:
        """
        Iterates over lines in the file, yielding each line as a string.

        Yields:
            str: Each line in the file, stripped of whitespace.
        """
        with zopen(self.filename, mode="rt", encoding="utf-8") as file:
            for line in file:
                yield cast("str", line.strip())

    def as_dict(self) -> dict[str, Any]:
        """
        Serializes the LobsterFile object to a dictionary.
        Spin keys in dictionaries are converted to strings for JSON compatibility.

        Returns:
            dict[str, Any]: Dictionary with keys "@module", "@class", "@version", and all attributes of the object.
        """
        dictionary = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "@version": None,
        }

        for k, v in vars(self).items():
            dictionary[k] = convert_spin_keys(v)

        return dictionary

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        """
        Deserializes a LobsterFile object from a dictionary. Spin keys in dictionaries are restored from strings.

        Args:
            d (dict[str, Any]): Dictionary produced by as_dict or similar serialization.

        Returns:
            LobsterFile: Instance of LobsterFile with attributes populated from the dictionary.
        """
        instance = cls.__new__(cls)

        decoded_dictionary = {
            k: restore_spin_keys(MontyDecoder().process_decoded(v)) for k, v in d.items() if not k.startswith("@")
        }

        for k, v in decoded_dictionary.items():
            setattr(instance, k, v)

        instance.filename = Path(instance.filename)

        return instance

    @property
    def has_spin(self) -> bool:
        """
        Indicates whether the file could contain spin-polarized data.

        Returns:
            bool: True if this file type supports spin, False otherwise.
        """
        return self.spins is not None

    @property
    def is_spin_polarized(self) -> bool:
        """
        Indicates whether the file contains spin-polarized data.

        Returns:
            bool: True if multiple spins are present, False otherwise.
        """
        return self.has_spin and len(self.spins) > 1


class LobsterInteractionsHolder(LobsterFile):
    """
    Container for LOBSTER interaction data. This class holds interaction metadata. It provides methods for filtering and
    retrieving interaction data based on various criteria.

    Attributes:
        interactions (list[LobsterInteractionData]): List of interaction metadata dicts.
    """

    interactions: list[LobsterInteractionData]
    data: NDArray[np.float64]

    def get_interaction_indices_by_properties(
        self,
        indices: list[int] | None = None,
        centers: list[str] | None = None,
        cells: list[list[int]] | None = None,
        orbitals: list[str] | None = None,
        length: tuple[float, float] | None = None,
    ) -> list[int]:
        """
        Returns indices of interactions that match provided property filters.

        Args:
            indices (list[int] | None): Optional sequence of interaction indices to match.
            centers (list[str] | None): Optional sequence of center name substrings; interaction must contain each
            center substring the required number of times.
            cells (list[list[int]] | None): Optional sequence of cell vectors to match against interaction cells.
            orbitals (list[str] | None): Optional sequence of orbital name substrings; interaction must contain each
            orbital substring the required number of times.
            length (tuple[float, float] | None): Optional (min, max) tuple to filter interactions by length inclusive.

        Returns:
            list[int]: Sorted list of interaction indices that match all supplied filters. If no filters are supplied,
            returns an empty list.
        """
        matching_sets = []

        if indices is not None:
            matching_sets.append({i for i, bond in enumerate(self.interactions) if bond["index"] in indices})

        if centers is not None:
            center_counts = Counter(centers)
            matching_centers = set()

            for i, bond in enumerate(self.interactions):
                bond_centers = bond.get("centers", [])

                if all(
                    sum(center_suffix in b for b in bond_centers) >= required_count
                    for center_suffix, required_count in center_counts.items()
                ):
                    matching_centers.add(i)

            matching_sets.append(matching_centers)

        if cells is not None:
            matching_sets.append(
                {
                    i
                    for i, bond in enumerate(self.interactions)
                    if bond.get("cells")
                    and all(np.all(np.equal(bond.get("cells"), cell), axis=1).any() for cell in cells if cell)
                }
            )

        if orbitals is not None:
            if not orbitals:
                matching_sets.append(
                    {i for i, bond in enumerate(self.interactions) if all(o is None for o in bond.get("orbitals", []))}
                )
            else:
                orbital_counts = Counter(orbitals)
                matching_orbitals = set()

                for i, bond in enumerate(self.interactions):
                    bond_orbitals = bond.get("orbitals", [])

                    if all(
                        sum(orbital_suffix in b for b in bond_orbitals if b) >= required_count
                        for orbital_suffix, required_count in orbital_counts.items()
                    ):
                        matching_orbitals.add(i)

                matching_sets.append(matching_orbitals)

        if length is not None:
            matching_sets.append(
                {
                    i
                    for i, bond in enumerate(self.interactions)
                    if (this_length := bond.get("length")) is not None and length[0] <= this_length <= length[1]
                }
            )

        return sorted(set.intersection(*matching_sets)) if matching_sets else []

    def get_interactions_by_properties(
        self: LobsterInteractionsHolder,
        indices: list[int] | None = None,
        centers: list[str] | None = None,
        cells: list[list[int]] | None = None,
        orbitals: list[str] | None = None,
        length: tuple[float, float] | None = None,
    ) -> list[LobsterInteractionData]:
        """Return interaction metadata dicts that match the provided filters.

        Args:
            indices (list[int] | None): Interaction indices to filter.
            centers (list[str] | None): Atom centers to filter.
            cells (list[list[int]] | None): Unit cell indices to filter.
            orbitals (list[str] | None): Orbitals to filter.
            length (tuple[float, float] | None): Length range to filter.

        Returns:
            list[dict]: List of interaction dictionaries matching the filters.
        """
        interaction_indices = self.get_interaction_indices_by_properties(indices, centers, cells, orbitals, length)

        return [bond for i, bond in enumerate(self.interactions) if i in interaction_indices]

    @staticmethod
    def get_label_from_interaction(
        interaction: LobsterInteraction,
        include_centers: bool = True,
        include_orbitals: bool = True,
        include_cells: bool = False,
        include_length: bool = False,
    ) -> str:
        """
        Generates a label string for a given interaction.

        Args:
            interaction (LobsterInteraction): Interaction metadata dictionary.
        Returns:
            str: Formatted label string representing the interaction.
        """
        parts = []

        for center, orbital, cell in zip(
            interaction["centers"],
            interaction["orbitals"],
            interaction["cells"],
            strict=True,
        ):
            tmp = ""
            if include_centers:
                tmp += center

            if include_cells and cell is not None:
                tmp += f"[{' '.join(map(str, cell))}]"

            if include_orbitals and orbital:
                tmp += f"[{orbital}]"

            parts.append(tmp)

        if not parts:
            raise ValueError(f"Could not generate label from interaction {interaction}")

        if include_length and interaction["length"] is not None:
            parts[-1] += f"({interaction['length']:.3f})"

        return "->".join(parts)

    @abstractmethod
    def process_data_into_interactions(self) -> None:
        """
        Abstract method to process raw data into structured interaction metadata. Must be implemented by subclasses.
        This method should populate the `interactions` attribute based on the raw `data` and any other relevant
        attributes. The exact processing logic will depend on the specific format and content of the data in the
        subclass.
        """
        ...

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        """Deserialize object from dictionary produced by `as_dict`.

        Args:
            d (dict): Dictionary produced by `as_dict`.

        Returns:
            COXXCAR: Reconstructed instance.
        """
        instance = super().from_dict(d)
        instance.data = np.asarray(instance.data, dtype=np.float64)

        instance.process_data_into_interactions()

        return instance



# cython: language_level=3
# cython: initializedcheck=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: cdivision=False
# cython: profile=False
# distutils: language = c

# Setting cdivision=True gives a small speed improvement, but the code is currently
# written based on Python division so using cdivision may result in missing neighbors
# in some off cases. See https://github.com/materialsproject/pymatgen/issues/2226

import numpy as np

cimport numpy as np
from libc.math cimport ceil, floor, pi, sqrt
from libc.stdlib cimport free, malloc, realloc
from libc.string cimport memset


cdef void *safe_malloc(size_t size) except? NULL:
    """Raise MemoryError if malloc fails."""
    if size == 0:
        return NULL
    cdef void *ptr = malloc(size)
    if ptr == NULL:
        raise MemoryError(f"Memory allocation of {size} bytes failed!")
    return ptr


cdef void *safe_realloc(void *ptr_orig, size_t size) except? NULL:
    """Raise MemoryError if realloc fails."""
    if size == 0:
        return NULL
    cdef void *ptr = realloc(ptr_orig, size)
    if ptr == NULL:
        raise MemoryError(f"Realloc memory of {size} bytes failed!")
    return ptr


def find_points_in_spheres(
        const double[:, ::1] all_coords,
        const double[:, ::1] center_coords,
        const double r,
        const np.int64_t[::1] pbc,
        const double[:, ::1] lattice,
        const double tol=1e-8,
        const double min_r=1.0
    ):
    """For each point in `center_coords`, get all the neighboring points in `all_coords`
    that are within the cutoff radius `r`. All the coordinates should be Cartesian.

    Args:
        all_coords: (np.ndarray[double, dim=2]) all available points.
            When periodic boundary is considered, this is all the points in the lattice.
        center_coords: (np.ndarray[double, dim=2]) all centering points
        r: (float) cutoff radius
        pbc: (np.ndarray[np.int64_t, dim=1]) whether to set periodic boundaries
        lattice: (np.ndarray[double, dim=2]) 3x3 lattice matrix
        tol: (float) numerical tolerance
        min_r: (float) minimal cutoff to calculate the neighbor list
            directly. If the cutoff is less than this value, the algorithm
            will calculate neighbor list using min_r as cutoff and discard
            those that have larger distances.

    Returns:
        index1 (n, ): Indexes of center_coords.
        index2 (n, ): Indexes of all_coords that form the neighbor pair.
        offset_vectors (n, 3): The periodic image offsets for all_coords.
        distances (n, ).
    """
    if r < min_r:
        findex1, findex2, foffset_vectors, fdistances = find_points_in_spheres(
            all_coords=all_coords, center_coords=center_coords,
            r=min_r + tol, pbc=pbc, lattice=lattice, tol=tol, min_r=min_r)
        mask = fdistances <= r
        return findex1[mask], findex2[mask], foffset_vectors[mask], fdistances[
            mask]

    cdef:
        int i, j, k
        unsigned int i_dim  # index of dimension
        unsigned int i_pt  # index of point in all_coords (n_total)
        double[3] max_rep  # maximum repetitions in each direction

        # Valid boundary, that is the minimum in center_coords - (r + tol)
        double[3] valid_min
        double[3] valid_max

        double ledge

        unsigned int n_center = center_coords.shape[0]
        unsigned int n_total = all_coords.shape[0]
        np.int64_t nlattice = 1

        np.int64_t[3] max_bounds = [1, 1, 1]
        np.int64_t[3] min_bounds = [0, 0, 0]
        double [:, ::1] frac_coords =  <double[:n_center, :3]> safe_malloc(
            n_center * 3 * sizeof(double)
        )
        double[:, ::1] all_frac_coords = <double[:n_total, :3]> safe_malloc(
            n_total * 3 * sizeof(double)
        )
        double[:, ::1] coords_in_cell = <double[:n_total, :3]> safe_malloc(
            n_total * 3 * sizeof(double)
        )
        double[:, ::1] offset_correction = <double[:n_total, :3]> safe_malloc(
            n_total * 3 * sizeof(double)
        )
        double[3][3] inv_lattice_arr
        double[:, ::1] inv_lattice = inv_lattice_arr
        double[3][3] reciprocal_lattice_arr = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        double[:, ::1] reciprocal_lattice = reciprocal_lattice_arr

        unsigned int count = 0
        unsigned int n_atoms = n_total
        double *offsets_p_temp = <double*> safe_malloc(n_atoms * 3 * sizeof(double))
        double *expanded_coords_p_temp = <double*> safe_malloc(
            n_atoms * 3 * sizeof(double)
        )
        np.int64_t *indices_p_temp = <np.int64_t*> safe_malloc(n_atoms * sizeof(np.int64_t))
        double coord_temp[3]
        np.int64_t ncube[3]

        np.int64_t[:, ::1] center_indices3 = <np.int64_t[:n_center, :3]> safe_malloc(
            n_center*3*sizeof(np.int64_t)
        )
        np.int64_t[::1] center_indices1 = <np.int64_t[:n_center]> safe_malloc(n_center*sizeof(np.int64_t))

        unsigned int malloc_chunk = 10000  # size of memory chunks to re-allocate dynamically
        unsigned int failed_malloc = 0  # flag for failed reallocation within loops
        np.int64_t *index_1 = <np.int64_t*> safe_malloc(malloc_chunk*sizeof(np.int64_t))
        np.int64_t *index_2 = <np.int64_t*> safe_malloc(malloc_chunk*sizeof(np.int64_t))
        double *offset_final = <double*> safe_malloc(3*malloc_chunk*sizeof(double))
        double *distances = <double*> safe_malloc(malloc_chunk*sizeof(double))
        np.int64_t cube_index_temp
        np.int64_t link_index
        double d_temp2
        double r2 = r * r

    if r < 0.1:
        ledge = 0.1
    else:
        ledge = r

    get_max_and_min(center_coords, valid_max, valid_min)
    for i_dim in range(3):
        valid_max[i_dim] = valid_max[i_dim] + r + tol
        valid_min[i_dim] = valid_min[i_dim] - r - tol

    # Process PBC
    get_frac_coords(lattice, inv_lattice, all_coords, offset_correction)
    for i_pt in range(n_total):
        for i_dim in range(3):
            if pbc[i_dim]:
                # Only wrap atoms when this dimension is PBC
                all_frac_coords[i_pt, i_dim] = offset_correction[i_pt, i_dim] % 1
                offset_correction[i_pt, i_dim] = offset_correction[i_pt, i_dim] - all_frac_coords[i_pt, i_dim]
            else:
                all_frac_coords[i_pt, i_dim] = offset_correction[i_pt, i_dim]
                offset_correction[i_pt, i_dim] = 0

    # Compute the reciprocal lattice in place
    get_reciprocal_lattice(lattice, reciprocal_lattice)

    get_max_rep(reciprocal_lattice, max_rep, r)

    # Get fractional coordinates of center points in place
    get_frac_coords(lattice, inv_lattice, center_coords, frac_coords)

    get_bounds(frac_coords, max_rep, &pbc[0], max_bounds, min_bounds)

    for i_dim in range(3):
        nlattice *= (max_bounds[i_dim] - min_bounds[i_dim])
    matmul(all_frac_coords, lattice, coords_in_cell)

    # Get translated images, coordinates and indices
    for i in range(min_bounds[0], max_bounds[0]):
        for j in range(min_bounds[1], max_bounds[1]):
            for k in range(min_bounds[2], max_bounds[2]):
                for i_pt in range(n_total):
                    for i_dim in range(3):
                        coord_temp[i_dim] = <double>i * lattice[0, i_dim] + \
                                        <double>j * lattice[1, i_dim] + \
                                        <double>k * lattice[2, i_dim] + \
                                        coords_in_cell[i_pt, i_dim]
                    if (
                            (coord_temp[0] > valid_min[0]) &
                            (coord_temp[0] < valid_max[0]) &
                            (coord_temp[1] > valid_min[1]) &
                            (coord_temp[1] < valid_max[1]) &
                            (coord_temp[2] > valid_min[2]) &
                            (coord_temp[2] < valid_max[2])
                    ):
                        offsets_p_temp[3*count] = i
                        offsets_p_temp[3*count+1] = j
                        offsets_p_temp[3*count+2] = k
                        indices_p_temp[count] = i_pt
                        expanded_coords_p_temp[3*count] = coord_temp[0]
                        expanded_coords_p_temp[3*count+1] = coord_temp[1]
                        expanded_coords_p_temp[3*count+2] = coord_temp[2]
                        count += 1
                        if count >= n_atoms:  # exceeding current memory
                            n_atoms += n_atoms
                            offsets_p_temp = <double*> realloc(
                                offsets_p_temp, n_atoms * 3 * sizeof(double)
                            )
                            expanded_coords_p_temp = <double*> realloc(
                                expanded_coords_p_temp, n_atoms * 3 * sizeof(double)
                            )
                            indices_p_temp = <np.int64_t*> realloc(
                                indices_p_temp, n_atoms * sizeof(np.int64_t)
                            )
                        if (
                                offset_final is NULL or
                                expanded_coords_p_temp is NULL or
                                indices_p_temp is NULL
                        ):
                            failed_malloc = 1
                            break
                else:
                    continue
                break
            else:
                continue
            break
        else:
            continue
        break

    if failed_malloc:
        raise MemoryError("A realloc of memory of failed!")
    else:
        failed_malloc = 0

    # If no valid neighbors were found return empty
    if count == 0:
        free(&frac_coords[0, 0])
        free(&all_frac_coords[0, 0])
        free(&coords_in_cell[0, 0])
        free(&offset_correction[0, 0])
        free(&center_indices1[0])
        free(&center_indices3[0, 0])

        free(offsets_p_temp)
        free(expanded_coords_p_temp)
        free(indices_p_temp)

        free(index_1)
        free(index_2)
        free(offset_final)
        free(distances)

        return (np.array([], dtype=np.int64), np.array([], dtype=np.int64),
            np.array([[], [], []], dtype=float).T, np.array([], dtype=float))

    n_atoms = count
    cdef:
        # Delete those beyond (min_center_coords - r - tol, max_center_coords + r + tol)
        double *offsets_p = <double*> safe_realloc(
            offsets_p_temp, count * 3 * sizeof(double)
        )
        double *expanded_coords_p = <double*> safe_realloc(
            expanded_coords_p_temp, count * 3 * sizeof(double)
        )
        np.int64_t *indices_p = <np.int64_t*> safe_realloc(
            indices_p_temp, count * sizeof(np.int64_t)
        )

        double[:, ::1] offsets = <double[:count, :3]> offsets_p
        double[:, ::1] expanded_coords = <double[:count, :3]> expanded_coords_p
        np.int64_t[::1] indices = <np.int64_t[:count]> indices_p

        # Construct linked cell list
        np.int64_t[:, ::1] all_indices3 = <np.int64_t[:n_atoms, :3]> safe_malloc(
            n_atoms * 3 * sizeof(np.int64_t)
        )
        np.int64_t[::1] all_indices1 = <np.int64_t[:n_atoms]> safe_malloc(
            n_atoms * sizeof(np.int64_t)
        )

    for i_dim in range(3):
        ncube[i_dim] = <np.int64_t>(ceil((valid_max[i_dim] - valid_min[i_dim]) / ledge))

    compute_cube_index(expanded_coords, valid_min, ledge, all_indices3)
    three_to_one(all_indices3, ncube[1], ncube[2], all_indices1)

    cdef:
        np.int64_t nb_cubes = ncube[0] * ncube[1] * ncube[2]
        np.int64_t *head = <np.int64_t*> safe_malloc(nb_cubes*sizeof(np.int64_t))
        np.int64_t *atom_indices = <np.int64_t*> safe_malloc(n_atoms*sizeof(np.int64_t))
        np.int64_t[:, ::1] neighbor_map = <np.int64_t[:nb_cubes, :27]> safe_malloc(
            nb_cubes * 27 * sizeof(np.int64_t)
        )

    memset(<void*>head, -1, nb_cubes*sizeof(np.int64_t))
    memset(<void*>atom_indices, -1, n_atoms*sizeof(np.int64_t))

    get_cube_neighbors(ncube, neighbor_map)
    for i_pt in range(n_atoms):
        atom_indices[i_pt] = head[all_indices1[i_pt]]
        head[all_indices1[i_pt]] = i_pt

    # Get center atoms' cube indices
    compute_cube_index(center_coords, valid_min, ledge, center_indices3)
    three_to_one(center_indices3, ncube[1], ncube[2], center_indices1)

    count = 0
    for i in range(n_center):
        for j in range(27):
            if neighbor_map[center_indices1[i], j] == -1:
                continue
            cube_index_temp = neighbor_map[center_indices1[i], j]
            link_index = head[cube_index_temp]
            while link_index != -1:
                d_temp2 = distance2(expanded_coords, center_coords, link_index, i, 3)
                if d_temp2 < r2 + tol:
                    index_1[count] = i
                    index_2[count] = indices[link_index]
                    offset_final[3*count] = offsets[link_index, 0] - offset_correction[indices[link_index], 0]
                    offset_final[3*count + 1] = offsets[link_index, 1] - offset_correction[indices[link_index], 1]
                    offset_final[3*count + 2] = offsets[link_index, 2] - offset_correction[indices[link_index], 2]
                    distances[count] = sqrt(d_temp2)

                    count += 1
                    # Increasing the memory size by allocating incrementally
                    # malloc_chunk more memory locations, I found it 3x faster to do so
                    # compared to using vectors in cpp
                    if count >= malloc_chunk:
                        malloc_chunk += malloc_chunk  # double the size
                        index_1 = <np.int64_t*> realloc(index_1, malloc_chunk * sizeof(np.int64_t))
                        index_2 = <np.int64_t*> realloc(index_2, malloc_chunk*sizeof(np.int64_t))
                        offset_final = <double*> realloc(
                            offset_final, 3*malloc_chunk*sizeof(double)
                        )
                        distances = <double*> realloc(
                            distances, malloc_chunk*sizeof(double)
                        )
                        if (
                                index_1 is NULL or index_2 is NULL or
                                offset_final is NULL or distances is NULL
                        ):
                            failed_malloc = 1
                            break
                link_index = atom_indices[link_index]
            else:
                continue
            break
        else:
            continue
        break

    if failed_malloc:
        raise MemoryError("A realloc of memory of failed!")
    else:
        failed_malloc = 0

    if count == 0:
        py_index_1 = np.array([], dtype=np.int64)
        py_index_2 = np.array([], dtype=np.int64)
        py_offsets = np.array([[], [], []], dtype=float).T
        py_distances = np.array([], dtype=float)
    else:
        # Resize to the actual size
        index_1 = <np.int64_t*>safe_realloc(index_1, count * sizeof(np.int64_t))
        index_2 = <np.int64_t*>safe_realloc(index_2, count*sizeof(np.int64_t))
        offset_final = <double*>safe_realloc(offset_final, 3*count*sizeof(double))
        distances = <double*>safe_realloc(distances, count*sizeof(double))

        # Convert to python objects
        py_index_1 = np.array(<np.int64_t[:count]>index_1)
        py_index_2 = np.array(<np.int64_t[:count]>index_2)
        py_offsets = np.array(<double[:count, :3]>offset_final)
        py_distances = np.array(<double[:count]>distances)

    # Free allocated memory
    free(index_1)
    free(index_2)
    free(offset_final)
    free(distances)
    free(offsets_p)
    free(expanded_coords_p)
    free(indices_p)
    free(head)
    free(atom_indices)

    free(&offset_correction[0, 0])
    free(&frac_coords[0, 0])
    free(&all_frac_coords[0, 0])
    free(&coords_in_cell[0, 0])
    free(&center_indices1[0])
    free(&center_indices3[0, 0])
    free(&neighbor_map[0, 0])

    free(&all_indices3[0, 0])
    free(&all_indices1[0])

    return py_index_1, py_index_2, py_offsets, py_distances


cdef void get_cube_neighbors(
        np.int64_t[3] ncube,
        np.int64_t[:, ::1] neighbor_map
    ):
    """
    Get {cube_index: cube_neighbor_indices} map.
    """
    cdef:
        int i, j, k
        unsigned int count = 0
        np.int64_t ncubes = ncube[0] * ncube[1] * ncube[2]
        np.int64_t[::1] counts = <np.int64_t[:ncubes]> safe_malloc(ncubes * sizeof(np.int64_t))
        np.int64_t[:, ::1] cube_indices_3d = <np.int64_t[:ncubes, :3]> safe_malloc(
            ncubes*3*sizeof(np.int64_t)
        )
        np.int64_t[::1] cube_indices_1d = <np.int64_t[:ncubes]> safe_malloc(ncubes*sizeof(np.int64_t))

        # Creating the memviews of c-arrays once substantially improves speed
        # but for some reason it makes the runtime scaling with the number of
        # atoms worse
        np.int64_t[1][3] index3_arr
        np.int64_t[:, ::1] index3 = index3_arr
        np.int64_t[1] index1_arr
        np.int64_t[::1] index1 = index1_arr

        int n = 1
        np.int64_t ntotal = (2 * n + 1) * (2 * n + 1) * (2 * n + 1)
        np.int64_t[:, ::1] ovectors
        np.int64_t *ovectors_p = <np.int64_t *> safe_malloc(ntotal * 3 * sizeof(np.int64_t))
        unsigned int n_ovectors = compute_offset_vectors(ovectors_p, n)

    # Resize to the actual size
    ovectors_p = <np.int64_t *> safe_realloc(ovectors_p, n_ovectors * 3 * sizeof(np.int64_t))
    ovectors = <np.int64_t[:n_ovectors, :3]>ovectors_p

    memset(<void*>&neighbor_map[0, 0], -1, neighbor_map.shape[0] * 27 * sizeof(np.int64_t))

    for i in range(ncubes):
        counts[i] = 0

    for i in range(ncube[0]):
        for j in range(ncube[1]):
            for k in range(ncube[2]):
                cube_indices_3d[count, 0] = i
                cube_indices_3d[count, 1] = j
                cube_indices_3d[count, 2] = k
                count += 1

    three_to_one(cube_indices_3d, ncube[1], ncube[2], cube_indices_1d)

    for i in range(ncubes):
        for j in range(27):
            index3[0, 0] = ovectors[j, 0] + cube_indices_3d[i, 0]
            index3[0, 1] = ovectors[j, 1] + cube_indices_3d[i, 1]
            index3[0, 2] = ovectors[j, 2] + cube_indices_3d[i, 2]
            if (
                    (index3[0, 0] < ncube[0]) &
                    (index3[0, 0] >= 0) &
                    (index3[0, 1] < ncube[1]) &
                    (index3[0, 1] >= 0) &
                    (index3[0, 2] < ncube[2]) &
                    (index3[0, 2] >= 0)
            ):
                three_to_one(index3, ncube[1], ncube[2], index1)
                neighbor_map[i, counts[i]] = index1[0]
                counts[i] += 1

    free(&cube_indices_3d[0, 0])
    free(&cube_indices_1d[0])
    free(&counts[0])
    free(ovectors_p)


cdef int compute_offset_vectors(
        np.int64_t* ovectors,
        np.int64_t n
    ) nogil:
    cdef:
        int i, j, k, ind
        int count = 0
        double center[8][3] # center vertices coords
        double offset[8][3]  # offsetted vertices

    for i in range(2):
        for j in range(2):
            for k in range(2):
                ind = i * 4 + j * 2 + k
                center[ind][0] = i - 0.5
                center[ind][1] = j - 0.5
                center[ind][2] = k - 0.5

    for i in range(-n, n + 1):
        for j in range(-n, n + 1):
            for k in range(-n, n + 1):
                offset_cube(center, i, j, k, offset)
                if distance_vertices(center, offset, n):
                    ovectors[3*count] = i
                    ovectors[3*count + 1] = j
                    ovectors[3*count + 2] = k
                    count += 1

    return count


cdef double distance2(
        const double[:, ::1] m1,
        const double[:, ::1] m2,
        np.int64_t index1,
        np.int64_t index2,
        np.int64_t size
    ) nogil:
    """Faster way to compute the distance squared by not using slice
    but providing indices in each matrix.
    """
    cdef:
        int i
        double s = 0

    for i in range(size):
        s += (m1[index1, i] - m2[index2, i]) * (m1[index1, i] - m2[index2, i])
    return s


cdef void get_bounds(
        const double[:, ::1] frac_coords,
        const double[3] max_rep,
        const np.int64_t[3] pbc,
        np.int64_t[3] max_bounds,
        np.int64_t[3] min_bounds
    ) nogil:
    """
    Given the fractional coordinates and the number of repeation needed in each
    direction (max_rep), compute the translational bounds in each dimension.
    """
    cdef:
        int i_dim
        double[3] max_fcoords
        double[3] min_fcoords

    get_max_and_min(frac_coords, max_fcoords, min_fcoords)

    for i_dim in range(3):
        min_bounds[i_dim] = 0
        max_bounds[i_dim] = 1

    for i_dim in range(3):
        if pbc[i_dim]:
            min_bounds[i_dim] = <np.int64_t>(floor(min_fcoords[i_dim] - max_rep[i_dim] - 1e-8))
            max_bounds[i_dim] = <np.int64_t>(ceil(max_fcoords[i_dim] + max_rep[i_dim] + 1e-8))

cdef void get_frac_coords(
        const double[:, ::1] lattice,
        double[:, ::1] inv_lattice,
        const double[:, ::1] cart_coords,
        double[:, ::1] frac_coords
    ) nogil:
    """
    Compute the fractional coordinates.
    """
    matrix_inv(lattice, inv_lattice)
    matmul(cart_coords, inv_lattice, frac_coords)

cdef void matmul(
        const double[:, ::1] m1,
        const double[:, ::1] m2,
        double [:, ::1] out
    ) nogil:
    """
    Matrix multiplication.
    """
    cdef:
        int i, j, k
        int m = m1.shape[0], n = m1.shape[1], l = m2.shape[1]

    for i in range(m):
        for j in range(l):
            out[i, j] = 0
            for k in range(n):
                out[i, j] += m1[i, k] * m2[k, j]

cdef void matrix_inv(
        const double[:, ::1] matrix,
        double[:, ::1] inv
    ) nogil:
    """
    Matrix inversion.
    """
    cdef:
        int i, j
        double det = matrix_det(matrix)

    for i in range(3):
        for j in range(3):
            inv[i, j] = (matrix[(j+1)%3, (i+1)%3] * matrix[(j+2)%3, (i+2)%3] -
                matrix[(j+2)%3, (i+1)%3] * matrix[(j+1)%3, (i+2)%3]) / det

cdef double matrix_det(
        const double[:, ::1] matrix
    ) nogil:
    """
    Matrix determinant.
    """
    return (
        matrix[0, 0] * (matrix[1, 1] * matrix[2, 2] - matrix[1, 2] * matrix[2, 1]) +
        matrix[0, 1] * (matrix[1, 2] * matrix[2, 0] - matrix[1, 0] * matrix[2, 2]) +
        matrix[0, 2] * (matrix[1, 0] * matrix[2, 1] - matrix[1, 1] * matrix[2, 0])
    )

cdef void get_max_rep(
        const double[:, ::1] reciprocal_lattice,
        double[3] max_rep,
        double r
    ) nogil:
    """
    Get maximum repetition in each directions.
    """
    cdef:
        unsigned int i_dim
        double recp_len

    for i_dim in range(3):  # TODO: is it ever not 3x3 for our cases?
        recp_len = norm(reciprocal_lattice[i_dim, :])
        max_rep[i_dim] = ceil((r + 0.15) * recp_len / (2 * pi))

cdef void get_reciprocal_lattice(
        const double[:, ::1] lattice,
        double[:, ::1] reciprocal_lattice
    ) nogil:
    """
    Compute the reciprocal lattice.
    """
    cdef unsigned int i
    for i in range(3):
        recip_component(
            lattice[i, :], lattice[(i+1)%3, :],
            lattice[(i+2)%3, :],
            reciprocal_lattice[i, :]
        )

cdef void recip_component(
        const double[::1] a1,
        const double[::1] a2,
        const double[::1] a3,
        double[::1] out
    ) nogil:
    """
    Compute the reciprocal lattice vector.
    """
    cdef:
        unsigned int i
        double prod
        double ai_cross_aj[3]

    cross(&a2[0], &a3[0], ai_cross_aj)  # need to pass address of memviews
    prod = inner(&a1[0], ai_cross_aj)
    for i in range(3):
        out[i] = 2 * pi * ai_cross_aj[i] / prod

cdef double inner(
    const double[3] x,
    const double[3] y
    ) nogil:
    """
    Compute inner product of 3d vectors.
    """
    cdef:
        double sum = 0
        unsigned int i

    for i in range(3):
        sum += x[i] * y[i]
    return sum

cdef void cross(
        const double[3] x,
        const double[3] y,
        double[3] out
    ) nogil:
    """
    Cross product of vector x and y, output in out.
    """
    out[0] = x[1] * y[2] - x[2] * y[1]
    out[1] = x[2] * y[0] - x[0] * y[2]
    out[2] = x[0] * y[1] - x[1] * y[0]

cdef double norm(
        const double[::1] vec
    ) nogil:
    """
    Vector norm.
    """
    cdef:
        unsigned int i
        unsigned int n = vec.shape[0]
        double sum = 0

    for i in range(n):
        sum += vec[i] * vec[i]
    return sqrt(sum)

cdef void get_max_and_min(
        const double[:, ::1] coords,
        double[3] max_coords,
        double[3] min_coords
    ) nogil:
    """
    Compute the lower (min_coords) and upper (max_coords) boundaries along each dimension.
    """
    cdef:
        unsigned int i_dim, i_pt
        unsigned int n_points = coords.shape[0]
        unsigned int n_dims = coords.shape[1]

    for i_dim in range(n_dims):
        max_coords[i_dim] = coords[0, i_dim]
        min_coords[i_dim] = coords[0, i_dim]

    for i_dim in range(n_points):
        for i_pt in range(n_dims):
            if coords[i_dim, i_pt] >= max_coords[i_pt]:
                max_coords[i_pt] = coords[i_dim, i_pt]
            if coords[i_dim, i_pt] <= min_coords[i_pt]:
                min_coords[i_pt] = coords[i_dim, i_pt]

cdef void compute_cube_index(
        const double[:, ::1] coords,
        const double[3] global_min,
        double radius,
        np.int64_t[:, ::1] return_indices
    ) nogil:
    """
    Computes the cube indices for a set of coordinates based on radius.
    """
    cdef unsigned int i_pt, i_dim

    for i_pt in range(coords.shape[0]):
        for i_dim in range(coords.shape[1]):
            return_indices[i_pt, i_dim] = <np.int64_t>(
                floor((coords[i_pt, i_dim] - global_min[i_dim] + 1e-8) / radius)
            )

cdef void three_to_one(
        const np.int64_t[:, ::1] label3d,
        np.int64_t ny,
        np.int64_t nz,
        np.int64_t[::1] label1d
    ) nogil:
    """
    3D vector representation to 1D.
    """
    cdef:
        unsigned int i
        unsigned int n = label3d.shape[0]

    for i in range(n):
        label1d[i] = label3d[i, 0] * ny * nz + label3d[i, 1] * nz + label3d[i, 2]

cdef bint distance_vertices(
        const double[8][3] center,
        const double[8][3] off,
        double r
    ) nogil:
    cdef:
        unsigned int i, j
        double d2
        double r2 = r * r

    for i in range(8):
        for j in range(8):
            d2 = (center[i][0] - off[j][0]) * (center[i][0] - off[j][0]) + \
                 (center[i][1] - off[j][1]) * (center[i][1] - off[j][1]) + \
                 (center[i][2] - off[j][2]) * (center[i][2] - off[j][2])
            if d2 <= r2:
                return 1
    return 0

cdef void offset_cube(
        const double[8][3] center,
        np.int64_t n,
        np.int64_t m,
        np.int64_t l,
        const double[8][3] (&offsetted)
    ) nogil:
    cdef unsigned int i, j, k

    for i in range(2):
        for j in range(2):
            for k in range(2):
                ind = i * 4 + j * 2 + k
                offsetted[ind][0] = center[ind][0] + n
                offsetted[ind][1] = center[ind][1] + m
                offsetted[ind][2] = center[ind][2] + l



from __future__ import annotations

import re
from typing import Any

from pymatgen.electronic_structure.core import Spin
from pymatgen.io.lobster.future.constants import LOBSTER_ORBITALS


def natural_sort(list_to_sort: str) -> list[Any]:
    """Sort a list of strings in human order.

    This function sorts strings in a way that humans would expect,
    taking into account numerical values within the strings.

    Args:
        list_to_sort (str): List of strings to sort.

    Returns:
        list[Any]: Sorted list of strings in human order.

    Example:
        >>> natural_sort(["file10", "file2", "file1"])
        ['file1', 'file2', 'file10']
    """
    return [int(text) if text.isdigit() else text.lower() for text in re.split(r"(\d+)", list_to_sort)]


def parse_orbital_from_text(text: str) -> str | None:
    """Parse the orbital from a text string of an ICOXXLIST file.

    This function extracts the orbital information from a given text string.
    It checks for valid orbital patterns and returns the matched orbital.

    Args:
        text (str): Text string to parse the orbital from.

    Returns:
        str | None: Parsed orbital string if a valid orbital is found,
        otherwise None.

    Example:
        >>> parse_orbital_from_text("1s_2p_x")
        '2p_x'
    """
    parts = text.split("_")

    if len(parts) == 1:
        return None

    for orbital in LOBSTER_ORBITALS:
        if match := re.search(rf"\d+{re.escape(orbital)}", "_".join(parts[-2:])):
            return match.group(0)

    return parts[-1] if re.match(r"\d+[a-z]+", parts[-1]) else None


def convert_spin_keys(obj: Any) -> Any:
    """Convert an object to a JSON-compatible format recursively.

    This function ensures that the input object is converted into a format
    that can be serialized into JSON. It handles lists, tuples, dictionaries,
    and enums.

    Args:
        obj (Any): Input object to convert.

    Returns:
        Any: JSON-compatible representation of the input object.

    Example:
        >>> make_json_compatible({"key": Enum("Example", "value")})
        {'key': 'value'}
    """
    if isinstance(obj, (list, tuple)):
        return [convert_spin_keys(item) for item in obj]

    if isinstance(obj, dict):
        new_dict = {}

        for k, v in obj.items():
            new_key = f"@Spin({k})" if isinstance(k, Spin) else k
            new_dict[new_key] = convert_spin_keys(v)

        return new_dict

    return obj


def restore_spin_keys(obj: Any) -> Any:
    """Restore Spin enum keys from JSON-serialized format.

    This function recursively processes an object and converts string keys
    "1" and "-1" back to Spin.up and Spin.down enum values. It handles
    nested lists, tuples, and dictionaries.

    Args:
        obj (Any): Input object with potential Spin keys as strings.

    Returns:
        Any: Object with Spin enum keys restored.

    Example:
        >>> restore_spin_keys({"1": [1, 2], "-1": [3, 4]})
        {<Spin.up: 1>: [1, 2], <Spin.down: -1>: [3, 4]}
    """
    if isinstance(obj, (list, tuple)):
        return type(obj)(restore_spin_keys(item) for item in obj)

    if isinstance(obj, dict):
        new_dict = {}
        for k, v in obj.items():
            if k == "@Spin(1)":
                new_key = Spin.up
            elif k == "@Spin(-1)":
                new_key = Spin.down
            else:
                new_key = k

            new_dict[new_key] = restore_spin_keys(v)

        return new_dict

    return obj



"""This module provides classes to handle thermal displacement matrices (anisotropic displacement parameters)."""

from __future__ import annotations

import re
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
from monty.json import MSONable

from pymatgen.core.structure import Structure
from pymatgen.core.structure_matcher import StructureMatcher
from pymatgen.io.cif import CifFile, CifParser, CifWriter, str2float
from pymatgen.symmetry.groups import SYMM_DATA
from pymatgen.util.due import Doi, due

if TYPE_CHECKING:
    from typing import Literal, Self

    from numpy.typing import ArrayLike, NDArray

    from pymatgen.util.typing import PathLike

__author__ = "J. George"
__copyright__ = "Copyright 2022, The Materials Project"
__version__ = "0.1"
__maintainer__ = "J. George"
__email__ = "janine.george@bam.de"
__status__ = "Testing"
__date__ = "August 09, 2022"

sub_spgrp = partial(re.sub, r"[\s_]", "")

space_groups = {sub_spgrp(k): k for k in SYMM_DATA["space_group_encoding"]}


class ThermalDisplacementMatrices(MSONable):
    """Handle thermal displacement matrices
    This class stores thermal displacement matrices in Ucart format.

    An earlier implementation based on Matlab can be found here:
    https://github.com/JaGeo/MolecularToolbox
    ( J. George, A. Wang, V. L. Deringer, R. Wang, R. Dronskowski, U. Englert, CrystEngComm, 2015, 17, 7414-7422.)
    """

    def __init__(
        self,
        thermal_displacement_matrix_cart: ArrayLike,
        structure: Structure,
        temperature: float | None,
        thermal_displacement_matrix_cif: ArrayLike | None = None,
    ) -> None:
        """
        Args:
            thermal_displacement_matrix_cart: 2D numpy array including the thermal_displacement matrix Ucart
                1st dimension atom types, then compressed thermal displacement matrix will follow
                U11, U22, U33, U23, U13, U12 (xx, yy, zz, yz, xz, xy)
                convention similar to "thermal_displacement_matrices.yaml" in phonopy
            structure: A pymatgen Structure object
            temperature: temperature at which thermal displacement matrix was determined
            thermal_displacement_matrix_cif: 2D numpy array including the thermal_displacement matrix Ucif format
                1st dimension atom types, then compressed thermal displacement matrix will follow
                U11, U22, U33, U23, U13, U12 (xx, yy, zz, yz, xz, xy)
                convention similar to "thermal_displacement_matrices.yaml" in phonopy.
        """
        self.thermal_displacement_matrix_cart = np.asarray(thermal_displacement_matrix_cart)
        self.structure = structure
        self.temperature = temperature
        if thermal_displacement_matrix_cif is not None:
            self.thermal_displacement_matrix_cif: None | NDArray = np.asarray(thermal_displacement_matrix_cif)
        else:
            self.thermal_displacement_matrix_cif = None

        # convert to matrix form easier handling of the data
        self.thermal_displacement_matrix_cart_matrixform = ThermalDisplacementMatrices.get_full_matrix(
            self.thermal_displacement_matrix_cart
        )

        if self.thermal_displacement_matrix_cif is not None:
            self.thermal_displacement_matrix_cif_matrixform = ThermalDisplacementMatrices.get_full_matrix(
                self.thermal_displacement_matrix_cif
            )

    @staticmethod
    def get_full_matrix(
        thermal_displacement: ArrayLike,
    ) -> np.ndarray:
        """Transfers the reduced matrix to the full matrix (order of reduced matrix U11, U22, U33, U23, U13, U12).

        Args:
            thermal_displacement: 2d numpy array, first dimension are the atoms

        Returns:
            3d numpy array including thermal displacements, first dimensions are the atoms
        """
        td = np.array(thermal_displacement)
        matrix_form = np.zeros((len(td), 3, 3))
        for idx, mat in enumerate(td):
            # xx, yy, zz, yz, xz, xy
            matrix_form[idx][0][0] = mat[0]
            matrix_form[idx][1][1] = mat[1]
            matrix_form[idx][2][2] = mat[2]
            matrix_form[idx][1][2] = mat[3]
            matrix_form[idx][2][1] = mat[3]
            matrix_form[idx][0][2] = mat[4]
            matrix_form[idx][2][0] = mat[4]
            matrix_form[idx][0][1] = mat[5]
            matrix_form[idx][1][0] = mat[5]
        return matrix_form

    @staticmethod
    def get_reduced_matrix(
        thermal_displacement: ArrayLike,
    ) -> np.ndarray:
        """Transfers the full matrix to reduced matrix (order of reduced matrix U11, U22, U33, U23, U13, U12).

        Args:
            thermal_displacement: 2d numpy array, first dimension are the atoms

        Returns:
            3d numpy array including thermal displacements, first dimensions are the atoms
        """
        td = np.array(thermal_displacement)
        reduced_matrix = np.zeros((len(td), 6))
        for idx, mat in enumerate(td):
            # xx, yy, zz, yz, xz, xy
            reduced_matrix[idx][0] = mat[0][0]
            reduced_matrix[idx][1] = mat[1][1]
            reduced_matrix[idx][2] = mat[2][2]
            reduced_matrix[idx][3] = mat[1][2]
            reduced_matrix[idx][4] = mat[0][2]
            reduced_matrix[idx][5] = mat[0][1]
        return reduced_matrix

    @property
    def Ustar(self) -> np.ndarray:
        """Computation as described in R. W. Grosse-Kunstleve, P. D. Adams, J Appl Cryst 2002, 35, 477-480.

        Returns:
            np.array: Ustar as array. First dimension are the atoms in the structure.
        """
        A = self.structure.lattice.matrix.T
        Ainv = np.linalg.inv(A)
        Ustar = []
        for mat in self.thermal_displacement_matrix_cart_matrixform:
            mat_ustar = np.dot(np.dot(Ainv, mat), Ainv.T)
            Ustar.append(mat_ustar)
        return np.array(Ustar)

    @property
    def Ucif(self) -> np.ndarray:
        """Computation as described in R. W. Grosse-Kunstleve, P. D. Adams, J Appl Cryst 2002, 35, 477-480.

        Returns:
            np.array: Ucif as array. First dimension are the atoms in the structure.
        """
        if self.thermal_displacement_matrix_cif is None:
            # computation as described in R. W. Grosse-Kunstleve, P. D. Adams, J Appl Cryst 2002, 35, 477-480.
            # will compute Ucif based on Ustar for each atom in the structure

            A = self.structure.lattice.matrix.T  # check this again?
            N = np.diag([np.linalg.norm(x) for x in np.linalg.inv(A)])
            Ninv = np.linalg.inv(N)
            Ucif = []
            Ustar = self.Ustar
            for mat in Ustar:
                mat_cif = np.dot(np.dot(Ninv, mat), Ninv.T)
                Ucif.append(mat_cif)
            return np.array(Ucif)
        return self.thermal_displacement_matrix_cif_matrixform

    @property
    def B(self) -> np.ndarray:
        """Computation as described in R. W. Grosse-Kunstleve, P. D. Adams, J Appl Cryst 2002, 35, 477-480.

        Returns:
            np.array: First dimension are the atoms in the structure.
        """
        B = []
        for mat in self.Ucif:
            mat_B = mat * 8 * np.pi**2
            B.append(mat_B)
        return np.array(B)

    @property
    def beta(self) -> list:
        """Computation as described in R. W. Grosse-Kunstleve, P. D. Adams, J Appl Cryst 2002, 35, 477-480.

        Returns:
            np.array: First dimension are the atoms in the structure.
        """
        # will compute beta based on Ustar
        beta = []
        for mat in self.Ustar:
            mat_beta = mat * 2 * np.pi**2
            beta.append(mat_beta)
        return beta

    @property
    def U1U2U3(self) -> list:
        """Computation as described in R. W. Grosse-Kunstleve, P. D. Adams, J Appl Cryst 2002, 35, 477-480.

        Returns:
            np.array: eigenvalues of Ucart. First dimension are the atoms in the structure.
        """
        thermal_disp_matrix = self.thermal_displacement_matrix_cart_matrixform
        return [np.linalg.eig(mat)[0] for mat in thermal_disp_matrix]

    def write_cif(self, filename: str) -> None:
        """Write a CIF including thermal displacements.

        Args:
            filename: name of the CIF file
        """
        writer = CifWriter(self.structure)
        writer.write_file(filename)
        # This will simply append the thermal displacement part to the CIF from the CifWriter
        # In the long run, CifWriter could be extended to handle thermal displacement matrices
        with open(filename, mode="a", encoding="utf-8") as file:
            file.write("loop_ \n")
            file.write("_atom_site_aniso_label\n")
            file.write("_atom_site_aniso_U_11\n")
            file.write("_atom_site_aniso_U_22\n")
            file.write("_atom_site_aniso_U_33\n")
            file.write("_atom_site_aniso_U_23\n")
            file.write("_atom_site_aniso_U_13\n")
            file.write("_atom_site_aniso_U_12\n")
            file.write(f"# Additional Data for U_Aniso: {self.temperature}\n")

            file.writelines(
                f"{site.specie.symbol}{idx} {matrix[0][0]} {matrix[1][1]} {matrix[2][2]}"
                f" {matrix[1][2]} {matrix[0][2]} {matrix[0][1]}\n"
                for idx, (site, matrix) in enumerate(zip(self.structure, self.Ucif, strict=True))
            )

    @staticmethod
    def _angle_dot(a: ArrayLike, b: ArrayLike) -> float:
        dot_product = np.dot(a, b)
        prod_of_norms = np.linalg.norm(a) * np.linalg.norm(b)
        divided = dot_product / prod_of_norms
        angle_rad = np.arccos(np.round(divided, 10))
        return np.degrees(angle_rad)

    @due.dcite(
        Doi("10.1039/C9CE00794F"),
        description="Angle: A new tool for validating theoretically derived anisotropic displacement "
        "parameters with experiment.",
    )
    def compute_directionality_quality_criterion(
        self, other: ThermalDisplacementMatrices
    ) -> list[dict[str, ArrayLike]]:
        """Will compute directionality of prolate displacement ellipsoids as described in
        https://doi.org/10.1039/C9CE00794F with the earlier implementation: https://github.com/damMroz/Angle/.

        Args:
            other: ThermalDisplacementMatrix
            please make sure that the order of the atoms in both objects that are compared
            is the same. Otherwise, this analysis will deliver wrong results

        Returns:
            will return a list including dicts for each atom that include "vector0"
            (largest principal axes of self object),
             "vector1" (largest principal axes of the other object), "angle" between both axes,
              These vectors can then, for example, be drawn into the structure with VESTA.
              Vectors are given in Cartesian coordinates
        """
        # compare the atoms string at least
        for spec1, spec2 in zip(self.structure.species, other.structure.species, strict=True):
            if spec1 != spec2:
                raise ValueError(
                    "Species in both structures are not the same! Please use structures that are similar to each other"
                )
        # check if structures match
        structure_match = StructureMatcher()
        if not structure_match.fit(struct1=self.structure, struct2=other.structure):
            raise ValueError("Structures have to be similar")

        results = []
        for self_Ucart, other_Ucart in zip(
            self.thermal_displacement_matrix_cart_matrixform,
            other.thermal_displacement_matrix_cart_matrixform,
            strict=True,
        ):
            result_dict = {}

            # determine eigenvalues and vectors for inverted Ucart
            invUcart_eig_self, invUcart_eigv_self = np.linalg.eig(np.linalg.inv(self_Ucart))
            invUcart_eig_other, invUcart_eigv_other = np.linalg.eig(np.linalg.inv(other_Ucart))

            argmin_self = np.argmin(invUcart_eig_self)
            vec_self = invUcart_eigv_self.transpose()[argmin_self]
            argmin_other = np.argmin(invUcart_eig_other)
            vec_other = invUcart_eigv_other.transpose()[argmin_other]
            # vector direction does not matter here, smallest angle should be given
            result_dict["angle"] = np.min(
                [
                    self._angle_dot(vec_self, vec_other),
                    self._angle_dot(vec_self, vec_other * -1),
                ]
            )
            result_dict["vector0"] = vec_self
            result_dict["vector1"] = vec_other

            results.append(result_dict)

        return results

    def visualize_directionality_quality_criterion(
        self,
        other: ThermalDisplacementMatrices,
        filename: PathLike = "visualization.vesta",
        which_structure: Literal[0, 1] = 0,
    ) -> None:
        """Will create a VESTA file for visualization of the directionality criterion.

        Args:
            other: ThermalDisplacementMatrices
            filename: Filename of the VESTA file
            which_structure: 0 means structure of the self object will be used,
                1 means structure of the other object will be used
        """
        # will return a VESTA file including vectors to visualize the quality criterion
        result = self.compute_directionality_quality_criterion(other=other)
        matrix_cif = (
            self.thermal_displacement_matrix_cif
            if self.thermal_displacement_matrix_cif is not None
            else self.get_reduced_matrix(self.Ucif)
        )

        if which_structure == 0:
            structure = self.structure
        elif which_structure == 1:
            structure = other.structure
        else:
            raise ValueError("Illegal which_structure value.")

        with open(filename, mode="w", encoding="utf-8") as file:
            file.write("#VESTA_FORMAT_VERSION 3.5.4\n \n \n")
            file.write("CRYSTAL\n\n")
            file.write("TITLE\n")
            file.write("Directionality Criterion\n\n")
            file.write("GROUP\n")
            file.write("1 1 P 1\n\n")
            file.write("CELLP\n")
            file.write(
                f"{structure.lattice.a} {structure.lattice.b} {structure.lattice.c} "
                f"{structure.lattice.alpha} {structure.lattice.beta} {structure.lattice.gamma}\n"
            )
            file.write("  0.000000   0.000000   0.000000   0.000000   0.000000   0.000000\n")  # error on parameters
            file.write("STRUC\n")  # codespell:ignore struc

            for site_idx, site in enumerate(structure, start=1):
                file.write(
                    f"{site_idx} {site.species_string} {site.species_string}{site_idx} 1.0000 {site.frac_coords[0]} "
                    f"{site.frac_coords[1]} {site.frac_coords[2]} 1a 1\n"
                )
                file.write(" 0.000000 0.000000 0.000000 0.00\n")  # error on positions - zero here

            # now we iterate over the whole structure and write down the fractional coordinates (with errors)
            file.write("  0 0 0 0 0 0 0\n")
            file.write("THERT 0\n")
            file.write("THERM\n")
            # print all U11s (make sure they are in the correct order)
            counter = 1
            # VESTA order: _U_12    _U_13    _atom_site_aniso_U_23
            for atom_therm, site in zip(matrix_cif, structure, strict=True):
                file.write(
                    f"{counter} {site.species_string}{counter} {atom_therm[0]} "
                    f"{atom_therm[1]} {atom_therm[2]} {atom_therm[5]} {atom_therm[4]} {atom_therm[3]}\n"
                )
                counter += 1
            file.write("  0 0 0 0 0 0 0 0\n")
            file.write("VECTR\n")
            vector_count = site_count = 1
            for vectors in result:
                vector0_x = vectors["vector0"][0]
                vector0_y = vectors["vector0"][1]
                vector0_z = vectors["vector0"][2]
                vector1_x = vectors["vector1"][0]
                vector1_y = vectors["vector1"][1]
                vector1_z = vectors["vector1"][2]

                file.write(f"    {vector_count} {vector0_x} {vector0_y} {vector0_z} 0\n")
                file.write(f"    {site_count} 0 0 0 0\n")

                file.write(" 0 0 0 0 0\n")
                vector_count += 1
                file.write(f"    {vector_count} {vector1_x} {vector1_y} {vector1_z} 0\n")
                file.write(f"    {site_count} 0 0 0 0\n")
                vector_count += 1
                site_count += 1
                file.write(" 0 0 0 0 0\n")

            file.write(" 0 0 0 0 0\n")
            file.write("VECTT\n")

            counter = 1
            # two vectors per atom
            for _ in range(len(result)):
                file.write(f"{counter} 0.2 255 0 0 1\n")
                counter += 1
                file.write(f"{counter} 0.2 0 0 255 1\n")
                counter += 1

            file.write(" 0 0 0 0 0\n")

    @property
    def ratio_prolate(self) -> np.ndarray:
        """This will compute ratio between largest and smallest eigenvalue of Ucart."""
        return np.array([np.max(us) / np.min(us) for us in self.U1U2U3])

    @classmethod
    def from_Ucif(
        cls,
        thermal_displacement_matrix_cif: ArrayLike,
        structure: Structure,
        temperature: float | None = None,
    ) -> Self:
        """Starting from a numpy array, it will convert Ucif values into Ucart values and initialize the class.

        Args:
            thermal_displacement_matrix_cif: np.array,
                first dimension are the atoms,
                then reduced form of thermal displacement matrix will follow
                Order as above: U11, U22, U33, U23, U13, U12
            structure: Structure object
            temperature: float
                Corresponding temperature

        Returns:
            ThermalDisplacementMatrices
        """
        # get matrix form
        thermal_displacement_matrix_cif_matrix_form = ThermalDisplacementMatrices.get_full_matrix(
            thermal_displacement_matrix_cif
        )

        # convert the parameters for each atom
        A = structure.lattice.matrix.T
        N = np.diag([np.linalg.norm(x) for x in np.linalg.inv(A)])
        Ucart = []
        for mat in thermal_displacement_matrix_cif_matrix_form:
            mat_ustar = np.dot(np.dot(N, mat), N.T)
            mat_ucart = np.dot(np.dot(A, mat_ustar), A.T)
            Ucart.append(mat_ucart)

        thermal_displacement_matrix_cart = ThermalDisplacementMatrices.get_reduced_matrix(np.array(Ucart))

        # get ThermalDisplacementMatrices Object

        return cls(
            thermal_displacement_matrix_cart=thermal_displacement_matrix_cart,
            thermal_displacement_matrix_cif=thermal_displacement_matrix_cif,
            structure=structure,
            temperature=temperature,
        )

    def to_structure_with_site_properties_Ucif(self) -> Structure:
        """Transfers this object into a structure with site properties (Ucif).
        This is useful for sorting the atoms in the structure including site properties.
        e.g. with code like this:
        def sort_order(site):
            return [site.specie.X, site.frac_coords[0], site.frac_coords[1], site.frac_coords[2]]
        new_structure0 = Structure.from_sites(sorted(structure0, key=sort_order)).

        Returns:
            Structure
        """
        site_properties: dict = {
            "U11_cif": [],
            "U22_cif": [],
            "U33_cif": [],
            "U23_cif": [],
            "U13_cif": [],
            "U12_cif": [],
        }
        if self.thermal_displacement_matrix_cif is None:
            cif_matrix = self.get_reduced_matrix(self.Ucif)
        else:
            cif_matrix = self.thermal_displacement_matrix_cif
        for atom_ucif in cif_matrix:
            # U11, U22, U33, U23, U13, U12
            site_properties["U11_cif"].append(atom_ucif[0])
            site_properties["U22_cif"].append(atom_ucif[1])
            site_properties["U33_cif"].append(atom_ucif[2])
            site_properties["U23_cif"].append(atom_ucif[3])
            site_properties["U13_cif"].append(atom_ucif[4])
            site_properties["U12_cif"].append(atom_ucif[5])

        return self.structure.copy(site_properties=site_properties)

    @classmethod
    def from_structure_with_site_properties_Ucif(cls, structure: Structure, temperature: float | None = None) -> Self:
        """Will create this object with the help of a structure with site properties.

        Args:
            structure: Structure object including U11_cif, U22_cif, U33_cif, U23_cif, U13_cif, U12_cif as site
            properties
            temperature: temperature for Ucif data

        Returns:
            ThermalDisplacementMatrices
        """
        # U11, U22, U33, U23, U13, U12
        Ucif_matrix = [[site.properties[f"U{idx}_cif"] for idx in (11, 22, 33, 23, 13, 12)] for site in structure]

        return cls.from_Ucif(Ucif_matrix, structure, temperature=temperature)

    @staticmethod
    def from_cif_P1(filename: str) -> list[ThermalDisplacementMatrices]:
        """Reads a CIF with P1 symmetry including positions and ADPs.
        Currently, no check of symmetry is performed as CifParser methods cannot be easily reused.

        Args:
            filename: Filename of the CIF.

        Returns:
            ThermalDisplacementMatrices
        """
        # This code is adapted from the CifParser
        # to reuse more code from there, the CifParser would need a major refactoring
        # Future plans: add this part to CifParser (i.e., include handling of symmetry for ADPs)

        cif = CifFile.from_file(filename)
        thermals = []
        for data in cif.data.values():
            # can only handle P1 symmetry for now

            lattice = CifParser.get_lattice_no_exception(data)

            all_coords = []
            all_species = []
            for idx in range(len(data["_atom_site_label"])):
                try:
                    # If site type symbol exists, use it. Otherwise, we use the
                    # label.
                    symbol = CifParser(filename)._parse_symbol(data["_atom_site_type_symbol"][idx])
                except KeyError:
                    symbol = CifParser(filename)._parse_symbol(data["_atom_site_label"][idx])
                if not symbol:
                    continue

                all_species.append(symbol)
                x = str2float(data["_atom_site_fract_x"][idx])
                y = str2float(data["_atom_site_fract_y"][idx])
                z = str2float(data["_atom_site_fract_z"][idx])

                all_coords.append([x, y, z])

            thermals_Ucif = [
                [
                    str2float(data["_atom_site_aniso_U_11"][idx]),
                    str2float(data["_atom_site_aniso_U_22"][idx]),
                    str2float(data["_atom_site_aniso_U_33"][idx]),
                    str2float(data["_atom_site_aniso_U_23"][idx]),
                    str2float(data["_atom_site_aniso_U_13"][idx]),
                    str2float(data["_atom_site_aniso_U_12"][idx]),
                ]
                for idx in range(len(data["_atom_site_aniso_label"]))
            ]
            struct = Structure(lattice, all_species, all_coords)

            thermal = ThermalDisplacementMatrices.from_Ucif(
                thermal_displacement_matrix_cif=thermals_Ucif,
                structure=struct,
                temperature=None,
            )
            thermals.append(thermal)
        return thermals



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypedDict

from numpy import complexfloating
from numpy.typing import NDArray

from pymatgen.electronic_structure.core import Spin

if TYPE_CHECKING:
    from typing import TypeAlias

    from numpy import floating, integer


class LobsterInteraction(TypedDict):
    """Dictionary representing a chemical interaction in LOBSTER.

    This dictionary stores information about a specific interaction between atoms in a structure.

    Attributes:
        index (int): The index of the interaction.
        centers (list[str]): List of strings representing the centers of the interaction (e.g., "Fe1", "O2").
        cells (list[list[int]] | NDArray[integer]): List of lists of integers representing the cells of the interaction
        (e.g., [0, 0, 0]).
        orbitals (list[str | None]): List of strings representing the orbitals involved in the interaction
        (e.g., "2s", "2p_x").
        length (float | None): The length of the interaction, representing the distance between the centers.
    """

    index: int
    centers: list[str]
    cells: list[list[int]] | NDArray[integer]
    orbitals: list[str | None]
    length: float | None


class LobsterInteractionData(LobsterInteraction, total=False):
    """Dictionary representing a chemical interaction in LOBSTER with additional COXX/ICOXX data.

    Extends `LobsterInteraction` by adding COXX and ICOXX values for each spin.

    Attributes:
        coxx (dict[Spin, NDArray[floating] | float]): COXX values for each spin.
        icoxx (dict[Spin, NDArray[floating] | float]): ICOXX values for each spin.
    """

    coxx: dict[Spin, NDArray[floating] | float]
    icoxx: dict[Spin, NDArray[floating] | float]


LobsterPopulations: TypeAlias = dict[str, dict[str, dict[Spin, dict[Literal["mulliken", "loewdin"], float]]]]

LobsterMatrixData: TypeAlias = dict[str, dict[Spin | None, NDArray[complexfloating]]]


class LobsterBandOverlaps(TypedDict):
    """Dictionary representing band overlaps in LOBSTER.

    Attributes:
        k_points (dict[Spin, list[list[float]]]): List of k-points for each spin.
        matrices (dict[Spin, list[NDArray[floating]]]): List of matrices for each spin.
        max_deviations (dict[Spin, list[float]]): List of maximal deviations for each spin.
    """

    k_points: dict[Spin, list[list[float]]]
    matrices: dict[Spin, list[NDArray[floating]]]
    max_deviations: dict[Spin, list[float]]


class LobsterFatband(TypedDict):
    """Dictionary representing fatband data in LOBSTER.

    Attributes:
        center (str): Atom associated with the fatband.
        orbital (str): Orbital associated with the fatband.
        energies (dict[Spin, NDArray[floating]]): Energies at each k-point for each spin.
        projections (dict[Spin, NDArray[floating]]): Weights/projections at each k-point for each spin.
    """

    center: str
    orbital: str
    energies: dict[Spin, NDArray[floating]]
    projections: dict[Spin, NDArray[floating]]


class LobsterFatbands(TypedDict):
    """Dictionary representing multiple fatbands in LOBSTER.

    Attributes:
        k_points (list[list[float]]): List of k-points.
        bands (list[LobsterFatband]): List of fatband dictionaries.
    """

    k_points: list[list[float]]
    bands: list[LobsterFatband]
